# Copyright 2014 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Impala identifiers

impala_identifiers = [
    'add',
    'aggregate',
    'all',
    'alter',
    'and',
    'api_version',
    'as',
    'asc',
    'avro',
    'between',
    'bigint',
    'binary',
    'boolean',
    'by',
    'cached',
    'case',
    'cast',
    'change',
    'char',
    'class',
    'close_fn',
    'column',
    'columns',
    'comment',
    'compute',
    'create',
    'cross',
    'data',
    'database',
    'databases',
    'date',
    'datetime',
    'decimal',
    'delimited',
    'desc',
    'describe',
    'distinct',
    'div',
    'double',
    'drop',
    'else',
    'end',
    'escaped',
    'exists',
    'explain',
    'external',
    'false',
    'fields',
    'fileformat',
    'finalize_fn',
    'first',
    'float',
    'format',
    'formatted',
    'from',
    'full',
    'function',
    'functions',
    'group',
    'having',
    'if',
    'in',
    'incremental',
    'init_fn',
    'inner',
    'inpath',
    'insert',
    'int',
    'integer',
    'intermediate',
    'interval',
    'into',
    'invalidate',
    'is',
    'join',
    'last',
    'left',
    'like',
    'limit',
    'lines',
    'load',
    'location',
    'merge_fn',
    'metadata',
    'not',
    'null',
    'nulls',
    'offset',
    'on',
    'or',
    'order',
    'outer',
    'overwrite',
    'parquet',
    'parquetfile',
    'partition',
    'partitioned',
    'partitions',
    'prepare_fn',
    'produced',
    'rcfile',
    'real',
    'refresh',
    'regexp',
    'rename',
    'replace',
    'returns',
    'right',
    'rlike',
    'row',
    'schema',
    'schemas',
    'select',
    'semi',
    'sequencefile',
    'serdeproperties',
    'serialize_fn',
    'set',
    'show',
    'smallint',
    'stats',
    'stored',
    'straight_join',
    'string',
    'symbol',
    'table',
    'tables',
    'tblproperties',
    'terminated',
    'textfile',
    'then',
    'timestamp',
    'tinyint',
    'to',
    'true',
    'uncached',
    'union',
    'update_fn',
    'use',
    'using',
    'values',
    'view',
    'when',
    'where',
    'with'
]
